/*
 * Copyright 2004-2011 Red Hat, Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 */

/* Needed before we include the kernel libdlm header */
#define __user
typedef uint8_t __u8;
typedef uint16_t __u16;
typedef uint32_t __u32;
#define BUILDING_LIBDLM


